////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2017 Josef Spidlen, Ph.D., FlowJo, LLC
//
// License
// The software is distributed under the terms of the 
// Artistic License 2.0
// http://www.r-project.org/Licenses/Artistic-2.0
// 
// Disclaimer
// This software and documentation come with no warranties of any kind.
// This software is provided "as is" and any express or implied 
// warranties, including, but not limited to, the implied warranties of
// merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the  copyright holder be liable for any direct, 
// indirect, incidental, special, exemplary, or consequential damages
// (including but not limited to, procurement of substitute goods or 
// services; loss of use, data or profits; or business interruption)
// however caused and on any theory of liability, whether in contract,
// strict liability, or tort arising in any way out of the use of this 
// software.    
//////////////////////////////////////////////////////////////////////////////

package com.flowjo.plugin.scimpute;
import java.awt.Component;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.swing.Box;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;

import com.flowjo.lib.parameters.ParameterSelectionPanel;
import com.flowjo.lib.parameters.ParameterSetMgrInterface;
import com.flowjo.lib.parameters.ParameterSelectionPanel.eParameterSelectionMode;
import com.treestar.flowjo.engine.EngineManager;
import com.treestar.lib.FJPluginHelper;
import com.treestar.lib.PluginHelper;
import com.treestar.lib.core.ExportFileTypes;
import com.treestar.lib.core.ExternalAlgorithmResults;
import com.treestar.lib.core.PopulationPluginInterface;
import com.treestar.lib.core.SeqGeqExternalAlgorithmResults;
import com.treestar.lib.data.StringUtil;
import com.treestar.lib.gui.FJButton;
import com.treestar.lib.gui.FontUtil;
import com.treestar.lib.gui.GuiFactory;
import com.treestar.lib.gui.HBox;
import com.treestar.lib.gui.actions.FJAction;
import com.treestar.lib.gui.numberfields.RangedDoubleTextField;
import com.treestar.lib.gui.numberfields.RangedIntegerTextField;
import com.treestar.lib.gui.panels.FJLabel;
import com.treestar.lib.xml.SElement;

public class scImpute implements PopulationPluginInterface, ActionListener {

	public static final String pluginVersion = "0.5";
	public static final String pluginName = "scImpute";
	
	public static final int fixedToolTipW = 300;
	public static final int fixedLabelW = 130;
	public static final int fixedLabelW2 = 400;
	public static final int fixedLabelH = 25;
	public static final int fixedFieldW = 75;
	public static final int fixedFieldH = 25;

	public static final int parSelPaneW = 400;
	public static final int parSelPaneH = 300;

	public static final String Failed = "Failed";
	public static final String NoRepeat = "Cannot repeat";
	public static final String cellIdParName = "CellId";
	
	protected static final String fShowOutputButtonLabel = "Show scImpute R script output";
	protected static final String fShowOutputButtonTooltip = "Show the execution of the scImpute R script";
	
    public static final String explainLabel    = "Select parameters for scImpute.";
    public static final String explainLabel2   = "Parameters used for scImpute.";
    public static final String explainTooltip  = "Single cell RNA-seq data is typically affected by so called dropouts - excess of zero or near zero counts due to the low amounts of mRNA within each individual cell. The scImpute plugin can impute the dropout values to improve the identification of cell types and further downsteam analysis.<br/>This is a selection of genes used for scImpute.";
    public static final String citationLabel   = "If you are using the scImpute plugin, please cite: <br/><font color=\"blue\">Wei Vivian Li, Jingyi Jessica Li. scImpute: Accurate And Robust Imputation For Single Cell RNA-Seq Data. BioRxiv preview. https://doi.org/10.1101/141598.</font>";
    public static final String reviewOutLabel  = "<br/>The plugin execution typically results in the creation of a new imputed data set, which is automatically added to your workspace. Here, you can review the execution of the underlying R script that performed the imputation.";
    
    public static final String clusterLabel = "Number of clusters";
    public static final String clusterTooltip = "Enter the number of clusters for cell similarity estimation.";
    public static final String thresholdLabel = "Dropout threshold";
    public static final String thresholdTooltip = "Dropout threshold as a value between 0 and 1; see paper for more details";
    public static final String coresLabel = "Computing cores (0 = all)";
    public static final String coresToolTip = "The number of computing cores used in second stage of imputing; 0 means all cores available will be used.";

    private static final String htmlLabelPref   = "<html><p width=\"" + fixedLabelW2 + "\">";
    private static final String htmlLabelSuff   = "</p></html>";
    private static final String htmlTooltipPref = "<html><p width=\"" + fixedToolTipW + "\">";
    private static final String htmlTooltipSuff = "</p></html>";
    
	protected static Icon myIcon = null;
	protected static final String sIconName = "images/impIcon.png";
	
	protected List<String> fParameterNames = new ArrayList<String>();
	
	protected int fK = -1;
	protected int fCores = 0;
	protected double fDropThre = 0.5;
	protected boolean fRanBefore = false;  // We only allow this plugin to run once, no recal, delete the node and try again if needed.
	protected String fRoutFile = "";
	
	protected RangedIntegerTextField fNumClustField;
	protected RangedIntegerTextField fNumCoresField;  
	protected RangedDoubleTextField  fNumDropoutThreshold;  
	
    @Override
	public String getName() 
	{
		return(pluginName);
	}

    @Override
	public String getVersion() 
	{
		return(pluginVersion);
	}
	
	@Override
	public Icon getIcon()
	{
		if (myIcon == null)
		{
			URL url = getClass().getClassLoader().getResource(getIconName());
			if (url != null)
				myIcon = new ImageIcon(url);
		}
		return myIcon;
	}
	
	protected String getIconName() { return sIconName; }
		
	@Override
	public ExportFileTypes useExportFileType() {
		return ExportFileTypes.CSV_PIR_SCALE;
	}

	public ExternalAlgorithmResults invokeAlgorithm(SElement fcmlQueryElement, File sampleFile, File outputFolder) 
	{
        SeqGeqExternalAlgorithmResults results = new SeqGeqExternalAlgorithmResults();
        
        if (fRanBefore) {
        	results.setWorkspaceString(scImpute.NoRepeat);
            return results;
        }
        fRanBefore = true;
        
        if(!sampleFile.exists()) 
        {
        	JOptionPane.showMessageDialog(null, "Input file did not exist", pluginName + " error", JOptionPane.ERROR_MESSAGE);
            results.setWorkspaceString(scImpute.Failed);
            return results;
        }
        
        if (fParameterNames.size() < 3)
		{
        	JOptionPane.showMessageDialog(null, "Not enough parameters were selected.", pluginName + " error", JOptionPane.ERROR_MESSAGE);
            results.setWorkspaceString(scImpute.Failed);
            return results;
		}
                
        String sampleName = sampleFile.getName();
		sampleName = sampleName.replaceAll(".FCS", "").replaceAll(".CSV", "").replaceAll(".ExtNode", "").replaceAll(".fcs", "").replaceAll(".LMD", "").replaceAll(".csv", "").trim();
        sampleName = StringUtil.rtrim(sampleName, ".");
        SCImputeCalc calculator = new SCImputeCalc();
        
        File resultCSV = calculator.runCalc(sampleFile, sampleName, fParameterNames, fK, fCores, fDropThre, outputFolder.getAbsolutePath());
        if (resultCSV != null && resultCSV.exists()) {
        	try {
        		File parentDir = resultCSV.getParentFile().getParentFile();
        		File movedResults = new File(parentDir, resultCSV.getName());
        		String[] fileNameAndExtension = parseFileName(resultCSV.getName());
        		int i = 1;
        		while (movedResults.exists() && i < 1000) {
        			movedResults = new File(parentDir, fileNameAndExtension[0] + " (" + i + ")" + fileNameAndExtension[1]);
        			i++;
        		}
        		if (movedResults.exists()) {
        			JOptionPane.showMessageDialog(null, "Failed to rename imputed results CSV file; file already exists.", pluginName + " error", JOptionPane.ERROR_MESSAGE);
                    results.setWorkspaceString(scImpute.Failed);
                    return results;
        		} else {
        			resultCSV.renameTo(movedResults);
        			FJPluginHelper.loadSamplesIntoWorkspace(fcmlQueryElement, new String[] { movedResults.getAbsolutePath().toString() } );
        		}
        	} catch (Exception e) {
        		JOptionPane.showMessageDialog(null, "Failed to move imputed results CSV file to the workspace directory.", pluginName + " error", JOptionPane.ERROR_MESSAGE);
                results.setWorkspaceString(scImpute.Failed);
                return results;
        	}        	
        } else {
        	JOptionPane.showMessageDialog(null, "scImpute R script failed to generate the results", pluginName + " error", JOptionPane.ERROR_MESSAGE);
            results.setWorkspaceString(scImpute.Failed);
            return results;
        }
        calculator.deleteScriptFile();
		File rOutFile = calculator.renameROutFile();
		if (rOutFile != null)
			fRoutFile = rOutFile.getAbsolutePath();

        results.setWorkspaceString("DONE!");
        return results;
	}

	private String[] parseFileName(String name)
	{
		String[] res = new String[2];
		res[0] = name.replaceFirst("[.][^.]+$", "");
		res[1] = "";
		int i = name.lastIndexOf('.');
		if (i > 0)
			res[1] = name.substring(i);
		return res;
	}

	@Override
	public SElement getElement()
	{
		SElement result = new SElement(getName());
		if (!fParameterNames.isEmpty())
		{
			SElement elem = new SElement(XMLConsts.Parameters);
			result.addContent(elem);
			for (String pName : fParameterNames)
			{
				SElement e = new SElement(XMLConsts.P);
				e.setString(XMLConsts.name, pName);
				elem.addContent(e);
			}
		}
		result.setInt(XMLConsts.K, fK);
		result.setInt(XMLConsts.cores, fCores);
		result.setDouble(XMLConsts.dropThre, fDropThre);
		result.setBool(XMLConsts.ranBefore, fRanBefore);
		result.setString(XMLConsts.routFile, fRoutFile);
		
		return result;	
	}

	@Override
	public void setElement(SElement element)
	{
		SElement params = element.getChild(XMLConsts.Parameters);
		if (params != null) {
			fParameterNames.clear();
			for (SElement elem : params.getChildren())
			{
				fParameterNames.add(elem.getString(XMLConsts.name));
			}			
		}
		fK = element.getInt(XMLConsts.K, fK);
		fCores = element.getInt(XMLConsts.cores, fCores);
		fDropThre = element.getDouble(XMLConsts.dropThre, fDropThre);
		fRanBefore = element.getBool(XMLConsts.ranBefore, fRanBefore);
		fRoutFile = element.getString(XMLConsts.routFile, fRoutFile);
	}

	@Override
	public List<String> getParameters()
	{
		return fParameterNames;
	}

	@Override
	public boolean promptForOptions(SElement fcmlQueryElement, List<String> pNames)
	{
		ParameterSetMgrInterface mgr = PluginHelper.getParameterSetMgr(fcmlQueryElement);
		if (mgr == null)
			return false;
		int nCells = getNumCellsInDataset(fcmlQueryElement);
		
		if (fK <= 0)
			fK = estimateK(nCells);
		
		List<Object> guiObjects = new ArrayList<Object>();
		
		FJLabel explainText = new FJLabel(explainLabel);
		if (fRanBefore) explainText.setText(explainLabel2);
		explainText.setToolTipText(htmlTooltipPref + explainTooltip + htmlTooltipSuff);
		guiObjects.add(new HBox(new Component[] { explainText }));
		ParameterSelectionPanel parSelPanel = null;
		
		if (fRanBefore) {			
			JList<String> jlistScroll = new JList<String>((String[])fParameterNames.toArray(new String[0]));
		    jlistScroll.setEnabled(false);
		    JScrollPane sp = new JScrollPane(jlistScroll);
		    Dimension dim = new Dimension(parSelPaneW, parSelPaneH / 2);
		    sp.setMaximumSize(dim);
			sp.setMinimumSize(dim);
			sp.setPreferredSize(dim);
			jlistScroll.setToolTipText(htmlTooltipPref + explainTooltip + htmlTooltipSuff);
			sp.setToolTipText(htmlTooltipPref + explainTooltip + htmlTooltipSuff);
		    guiObjects.add(sp);
		} else {
			parSelPanel = new ParameterSelectionPanel(mgr, eParameterSelectionMode.WithSetsAndParameters, true, false, false, true);
			parSelPanel.setSelectedParameters(fParameterNames);
			Dimension dim = new Dimension(parSelPaneW, parSelPaneH);
			parSelPanel.setMaximumSize(dim);
			parSelPanel.setMinimumSize(dim);
			parSelPanel.setPreferredSize(dim);
			guiObjects.add(parSelPanel);
		}
		
        FJLabel fjLabelNumClusters = new FJLabel(clusterLabel);
        fNumClustField = new RangedIntegerTextField(1, 100);
        fNumClustField.setInt(fK);
        fNumClustField.setToolTipText(htmlTooltipPref + clusterTooltip + htmlTooltipSuff);
        GuiFactory.setSizes(fNumClustField, new Dimension(fixedFieldW, fixedFieldH));
        GuiFactory.setSizes(fjLabelNumClusters, new Dimension(fixedLabelW, fixedLabelH));
        guiObjects.add(new HBox(new Component[] { fjLabelNumClusters, fNumClustField }));

        FJLabel fjLabelThreshold = new FJLabel(thresholdLabel);
        fNumDropoutThreshold = new RangedDoubleTextField(0, 1, new DoubleNumberFormatter());
        fNumDropoutThreshold.setDouble(fDropThre);
        fNumDropoutThreshold.setToolTipText(htmlTooltipPref + thresholdTooltip + htmlTooltipSuff);
        GuiFactory.setSizes(fNumDropoutThreshold, new Dimension(fixedFieldW, fixedFieldH));
        GuiFactory.setSizes(fjLabelThreshold, new Dimension(fixedLabelW, fixedLabelH));
        guiObjects.add(new HBox(new Component[] { fjLabelThreshold, fNumDropoutThreshold }));

        FJLabel fjLabelNumCores = new FJLabel(coresLabel);
        fNumCoresField = new RangedIntegerTextField(0, 32);
        fNumCoresField.setInt(fCores);
        fNumCoresField.setToolTipText(htmlTooltipPref + coresToolTip + htmlTooltipSuff);
        GuiFactory.setSizes(fNumCoresField, new Dimension(fixedFieldW, fixedFieldH));
        GuiFactory.setSizes(fjLabelNumCores, new Dimension(fixedLabelW, fixedLabelH));
        // More than one core is not supported by scImpute on Windows, let's not show that component on Windows
        // TODO: This is not ideal; one day the client part and the computing part may be on different machines (and different OS) 
        if(!EngineManager.isWindows())
        	guiObjects.add(new HBox(new Component[] { fjLabelNumCores, fNumCoresField }));
        
		FJLabel citationText = new FJLabel(htmlLabelPref + citationLabel + htmlLabelSuff);
		citationText.setToolTipText(htmlTooltipPref + citationLabel + htmlTooltipSuff);
		guiObjects.add(new HBox(new Component[] { citationText }));
		
		if (fRanBefore) {
			fNumClustField.setEnabled(false);
			fNumDropoutThreshold.setEnabled(false);
			fNumCoresField.setEnabled(false);
			if (fRoutFile != null && !fRoutFile.isEmpty()) {
				File rOut = new File(fRoutFile);
				if (rOut != null && rOut.exists()) {					
					FJLabel reviewOutText = new FJLabel(htmlLabelPref + reviewOutLabel + htmlLabelSuff);
					reviewOutText.setToolTipText(htmlTooltipPref + reviewOutLabel + htmlTooltipSuff);
					guiObjects.add(new HBox(new Component[] { reviewOutText }));
					guiObjects.add(new HBox(Box.createHorizontalGlue(),  createShowOutputButton(), Box.createHorizontalGlue()));
				}
			}			
			JOptionPane.showMessageDialog((Component) null, guiObjects.toArray(), getName() + " " + getVersion(), JOptionPane.PLAIN_MESSAGE);
			return false;
		} else {
			int option = JOptionPane.showConfirmDialog((Component) null, guiObjects.toArray(), getName() + " " + getVersion(), JOptionPane.OK_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE);
			if (option == JOptionPane.OK_OPTION)
			{
				fK = fNumClustField.getInt();
				fCores = fNumCoresField.getInt();
				fDropThre = fNumDropoutThreshold.getDouble();
				String warning = "";
				int confirmation = JOptionPane.OK_OPTION;
				if (nCells > 500)
				{
					String[] estimate = estimateTime(nCells);
					warning = "Please note the calculation will take a while, likely somewhere between " + estimate[0] + " and " + estimate[1] + ".\nAre you sure you want to do this?";
					confirmation = JOptionPane.showConfirmDialog((Component) null, warning, getName() + " " + getVersion(), JOptionPane.OK_CANCEL_OPTION);
				}
				if (confirmation != JOptionPane.OK_OPTION)
					return false;

				if(parSelPanel != null) {
					fParameterNames.clear();
					fParameterNames.addAll(parSelPanel.getParameterSelection());
				}
				return true;
			}
			else
				return false;
		}
	}

	private Component createShowOutputButton()
	{
		FJButton outButton = new FJButton(new FJAction(this, fShowOutputButtonLabel), FontUtil.dlog10);
		outButton.setToolTipText(fShowOutputButtonTooltip);
		return outButton;
	}

	public static int getNumCellsInDataset(SElement fcmlQueryElement) {
		int ret = 3;
		SElement FcmlQuery = fcmlQueryElement.getChild(XMLConsts.FcmlQuery);
		if (FcmlQuery == null) return ret;
		SElement DataSet = FcmlQuery.getChild(XMLConsts.DataSet);
		if (DataSet == null) return ret;
		int numEvents = DataSet.getInt(XMLConsts.numEvents);
		return numEvents;
	}
	
	public static int estimateK(int nCells) {
		if (nCells >= 1024)
			return (int)(1.5 * myLog(nCells, 2));
		if (nCells > 96)
			return (int)(2.0+(13.0*((double)nCells/1024.0)));
		if (nCells > 60)
			return 3;
		return 2;
	}
	
	public static double myLog(double x, double base) {
		return ((int)(Math.log(x) / Math.log(base)));
	}
	
	public static String[] estimateTime(int nCells) {
		String[] ret = new String[2];
		int lowSec = nCells / 2;
		int highSec = 5 * nCells;
		ret[0] = stringFromSec(lowSec);
		ret[1] = stringFromSec(highSec);
		return ret;
	}

	public static String stringFromSec(int n) {
		String unit = "second";
		if (n > 60) {
			n /= 60;
			unit = "minute";
		}
		if (n > 120) {
			n /= 60;
			unit = "hour";
		}
		if (n > 48) {
			n /= 24;
			unit = "hour";
		}
		if (n > 1) {
			unit = unit + "s";
		}
		return "" + n + " " + unit;
	}

	@Override
	public void actionPerformed(ActionEvent e)
	{
		try {
			if(fShowOutputButtonLabel.equals(e.getActionCommand()) && fRoutFile != null && !fRoutFile.isEmpty())
			{
				File file = new File(fRoutFile);
				if (file != null && file.exists())
					Desktop.getDesktop().open(file);
				else
					Desktop.getDesktop().browse(new URI(fRoutFile));
			}
		}
		catch (IOException e1) { }
		catch (URISyntaxException e1) { }
	}
}
